"""Wagtail Live utils"""

import re
from functools import lru_cache

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string
from wagtail.embeds.oembed_providers import all_providers

SUPPORTED_MIME_TYPES = ["png", "jpeg", "gif"]


def get_setting_or_raise(setting, setting_str):
    """
    Retrieves and returns the value of a setting if present in user's settings,
    else raises an `ImproperlyConfigured` error.

    Args:
        setting (str):
            The name of the setting to find in user's settings.
        setting_str (str):
            A verbose name for the setting when reporting errors.

    Returns:
        str: The value of the setting if defined in user's settings.

    Raises:
        ImproperlyConfigured: If the setting isn't defined.
    """

    value = getattr(settings, setting, "")
    if not value:
        err_msg = f"You haven't specified a {setting_str} in your settings."
        raise ImproperlyConfigured(err_msg)
    return value


def get_live_page_model():
    """
    Retrieves the live page model specified in user's settings.

    Returns:
        LivePageMixin:
            The live page model specified.

    Raises:
        ImproperlyConfigured: if no live page model is specified or
            the one specified doesn't inherit from `wagtail_live.models.LivePageMixin`.
        ImportError: if the live page model class couldn't be loaded.
    """

    from wagtail_live.models import LivePageMixin

    live_model_class = get_setting_or_raise(
        setting="WAGTAIL_LIVE_PAGE_MODEL", setting_str="live page model"
    )

    model = import_string(live_model_class)

    if not issubclass(model, LivePageMixin):
        raise ImproperlyConfigured(
            f"The live page model {live_model_class} doesn't inherit from "
            "wagtail_live.models.LivePageMixin."
        )

    return model


def get_live_receiver():
    """
    Retrieves the live receiver specified in user's settings.

    Returns:
        BaseMessageReceiver: The live receiver specified if found else `None`.

    Raises:
        ImproperlyConfigured: if the receiver specified doesn't inherit from
            `wagtail_live.receivers.BaseMessageReceiver`.
        ImportError: if the receiver class couldn't be loaded.
    """

    from wagtail_live.receivers.base import BaseMessageReceiver

    live_receiver_class = getattr(settings, "WAGTAIL_LIVE_RECEIVER", "")
    if not live_receiver_class:
        # Assume live interface is used, in which case no additional setup is needed.
        return

    receiver = import_string(live_receiver_class)

    if not issubclass(receiver, BaseMessageReceiver):
        raise ImproperlyConfigured(
            f"The receiver {live_receiver_class} doesn't inherit from "
            "wagtail_live.receivers.BaseMessageReceiver."
        )
    return receiver


def get_live_publisher():
    """
    Retrieves the live publisher specified in user's settings.

    Returns:
        Publisher: The live publisher specified.

    Raises:
        ImproperlyConfigured: if no publisher class is specified in settings.
        ImportError: if the publisher class couldn't be loaded.
    """

    live_publisher_class = get_setting_or_raise(
        setting="WAGTAIL_LIVE_PUBLISHER", setting_str="live publisher"
    )
    return import_string(live_publisher_class)


def get_polling_timeout():
    """
    Retrieves the duration of the polling timeout for the long polling technique.

    The user can set this parameter in his settings by doing so:
    ```python
    WAGTAIL_LIVE_POLLING_TIMEOUT = (duration in seconds)
    ```

    The default value is 60 seconds.

    Returns:
        int: The duration of the polling timeout if defined else 60.
    """

    return getattr(settings, "WAGTAIL_LIVE_POLLING_TIMEOUT", 60)


def get_polling_interval():
    """
    Retrieves the duration of the polling interval for the interval polling technique.

    The user can set this parameter in his settings by doing so:
    ```python
    WAGTAIL_LIVE_POLLING_INTERVAL = (duration in ms)
    ```
    The default value is 3000 milliseconds.

    Returns:
        int: the duration of the polling interval if defined else 3000.
    """

    return getattr(settings, "WAGTAIL_LIVE_POLLING_INTERVAL", 3000)


@lru_cache(maxsize=None)
def is_embed(text):
    """
    Checks if a text is a link to embed.

    Args:
        text (str): Text to check

    Returns:
        bool:
        - `True` if text corresponds to an embed link.
        - `False` else.
    """

    for provider in all_providers:
        for url_pattern in provider.get("urls", []):
            if bool(re.match(url_pattern, text)):
                return True

    return False



"""Wagtail Live URLs."""

import logging

from django.core.exceptions import ImproperlyConfigured

from wagtail_live.exceptions import WebhookSetupError
from wagtail_live.utils import get_live_publisher, get_live_receiver

logger = logging.getLogger(__name__)

urlpatterns = []

try:
    live_publisher = get_live_publisher()
    if hasattr(live_publisher, "get_urls"):
        urlpatterns += live_publisher.get_urls()
except ImproperlyConfigured as err:
    logger.error(f"{err}")

live_receiver = get_live_receiver()
if live_receiver:
    try:
        urlpatterns += live_receiver.get_urls()
    except WebhookSetupError:
        logger.error("Webhook setup failed")



"""Wagtail Live receiver classes."""

import json
import logging

from django.http import HttpResponse, HttpResponseForbidden
from django.urls import path
from django.utils.decorators import method_decorator
from django.utils.functional import cached_property
from django.utils.timezone import now
from django.views import View
from django.views.decorators.csrf import csrf_exempt
from wagtail.images import get_image_model

from wagtail_live.blocks import (
    add_block_to_live_post,
    clear_live_post_content,
    construct_embed_block,
    construct_image_block,
    construct_live_post_block,
    construct_text_block,
)
from wagtail_live.exceptions import RequestVerificationError
from wagtail_live.utils import SUPPORTED_MIME_TYPES, get_live_page_model, is_embed

logger = logging.getLogger(__name__)


TEXT = "text"
IMAGE = "image"
EMBED = "embed"
LivePost = "live_post"


class BaseMessageReceiver:
    """Base Receiver class."""

    @cached_property
    def model(self):
        """
        LivePageMixin is an abstract class, so we can't make queries directly
        We have to get the actual model which subclasses it to perform queries.
        """

        return get_live_page_model()

    def dispatch_event(self, event):
        """
        Dispatches an event to find the corresponding handler.

        Args:
            event (dict): New event from a messaging app.
        """

        raise NotImplementedError

    def get_channel_id_from_message(self, message):
        """
        Retrieves a channel ID from a message.

        Args:
            message (dict): A received message from a messaging app.

        Returns:
            str: ID of the channel which the given message belongs to.
        """

        raise NotImplementedError

    def get_live_page_from_channel_id(self, channel_id):
        """
        Retrieves the live page with a given channel ID.

        Args:
            channel_id (str): Channel ID

        Returns:
            LivePageMixin: The livepage corresponding to `channel_id`.

        Raises:
            Http404: if a page with the given `channel_id` doesn't exist.
        """

        return self.model.objects.get(channel_id=channel_id)

    def get_message_id_from_message(self, message):
        """
        Retrieves message's ID.

        Args:
            message (dict): A received message from a messaging app

        Returns:
            str: Id of the given message
        """

        raise NotImplementedError

    def get_message_text(self, message):
        """
        Retrieves the text of a message.

        A message is made of text and files.

        Args:
            message (dict): A received message from a messaging app

        Returns:
            str: Text of the given message.
        """

        raise NotImplementedError

    def get_message_files(self, message):
        """
        Retrieves the files of a message.

        A message is made of text and files.

        Args:
            message (dict): A received message from a messaging app

        Returns:
            list: Files included in the given message.
        """

        raise NotImplementedError

    def get_image_title(self, image):
        """
        Retrieves the title of an image.

        Args:
            image (dict): Informations about an image

        Returns:
            str: Title of the image.
        """

        raise NotImplementedError

    def get_image_name(self, image):
        """
        Retrieves the name of an image.

        Args:
            image (dict): Informations about an image

        Returns:
            str: Name of the image.
        """

        raise NotImplementedError

    def get_image_mimetype(self, image):
        """
        Retrieves the mimetype of an image.

        Args:
            image (dict): Informations about an image

        Returns:
            str: Mimetype of the image.
        """

        raise NotImplementedError

    def get_image_content(self, image):
        """
        Retrieves the content of an image.

        Args:
            image (dict): Informations about an image

        Returns:
            File: Content of the image.
        """

        raise NotImplementedError

    def get_image_dimensions(self, image):
        """
        Retrieves the width and height of an image.

        Args:
            image (dict): Informations about an image

        Returns:
            (int, int): Width and height of the image.

        Raises:
            ValueError: if the width and height of the image can't be retrieved.
        """

        raise NotImplementedError

    def get_message_id_from_edited_message(self, message):
        """
        Retrieves the ID of the original message.

        Args:
            message (dict): A received message from a messaging app.

        Returns:
            str: ID of the original message that is being edited.
        """

        raise NotImplementedError

    def get_message_text_from_edited_message(self, message):
        """
        Retrieves the text an edited message.

        Args:
            message (dict): A received message from a messaging app.

        Returns:
            str: Text of the edited message.
        """

        raise NotImplementedError

    def get_message_files_from_edited_message(self, message):
        """
        Retrieves the files from an edited message.

        Args:
            message (dict): A received message from a messaging app.

        Returns:
            list: Files of the edited message.
        """

        raise NotImplementedError

    def get_embed(self, text):
        """
        Checks if a text is an embed for this receiver and return embed URL if so.

        Args:
            text (str): Text to check.

        Returns:
            str: URL of the embed if the text contains an embed, else `""`.
        """

        return text if is_embed(text=text) else ""

    def parse_text(self, text):
        """
        Parses a raw text content according to the input source formatting rules.

        Args:
            text (str): a text.

        Returns:
            str: the actual content of the text. (Returns the text itself by default).
        """

        return text

    def process_text(self, live_post, message_text):
        """
        Processes the text of a message.

        Parses the message, constructs corresponding block types i.e Embed or Text
        and add those blocks to the given live post.

        Args:
            live_post (LivePostBlock):
                Live post to update.
            message_text (str):
                Text to add to a live post.
        """

        message_parts = message_text.split("\n")
        for text in message_parts:
            # Avoid creating a block for empty content
            text = text.strip()
            if not text:
                continue

            block_type = ""
            url = self.get_embed(text=text)
            if url:
                block = construct_embed_block(url=url)
                block_type = EMBED

            else:
                text_content = self.parse_text(text=text)
                block = construct_text_block(text=text_content)
                block_type = TEXT

            add_block_to_live_post(
                block_type=block_type,
                block=block,
                live_block=live_post,
            )

    def process_files(self, live_post, files):
        """
        Processes the files of a message.

        Creates the corresponding block for any file and adds it to the given live post.

        Args:
            live_post (LivePostBlock):
                Live post to update.
            files (list):
                Files to add to a live post.
        """

        for item in files:
            image_title = self.get_image_title(image=item)
            try:
                image_width, image_height = self.get_image_dimensions(image=item)
            except ValueError:
                logger.error(f"Unable to retrieve the dimensions of {image_title}")
                continue

            mime_type = self.get_image_mimetype(image=item)
            if mime_type not in SUPPORTED_MIME_TYPES:
                logger.error(
                    f"Couldn't upload {image_title}. "
                    + f"Images of type {mime_type} aren't supported yet."
                )
                continue

            image = get_image_model()(
                title=image_title,
                width=image_width,
                height=image_height,
            )

            image_content = self.get_image_content(image=item)
            image_name = self.get_image_name(image=item)
            image.file.save(name=image_name, content=image_content, save=True)

            block = construct_image_block(image=image)
            add_block_to_live_post(
                block_type=IMAGE,
                block=block,
                live_block=live_post,
            )

    def add_message(self, message):
        """
        Adds a received message from a messaging app to the live page corresponding
        to the channel where the message was posted if such a page exists.

        Args:
            message (dict): A message received from a messaging app.
        """

        channel_id = self.get_channel_id_from_message(message=message)
        try:
            live_page = self.get_live_page_from_channel_id(channel_id=channel_id)
        except self.model.DoesNotExist:
            return

        message_id = self.get_message_id_from_message(message=message)
        live_post = construct_live_post_block(message_id=message_id, created=now())

        message_text = self.get_message_text(message=message)
        self.process_text(live_post=live_post, message_text=message_text)

        files = self.get_message_files(message=message)
        self.process_files(live_post=live_post, files=files)

        live_page.add_live_post(live_post=live_post)

    def change_message(self, message):
        """
        Changes an edited message in a messaging app in the live page corresponding
        to the channel where the message was posted if such a page exists.

        Args:
            message (dict): A message edited from a messaging app.
        """

        channel_id = self.get_channel_id_from_message(message=message)
        try:
            live_page = self.get_live_page_from_channel_id(channel_id=channel_id)
        except self.model.DoesNotExist:
            return

        message_id = self.get_message_id_from_edited_message(message=message)
        live_post = live_page.get_live_post_by_message_id(message_id=message_id)
        clear_live_post_content(live_post=live_post)

        message_text = self.get_message_text_from_edited_message(message=message)
        self.process_text(live_post=live_post.value, message_text=message_text)

        files = self.get_message_files_from_edited_message(message=message)
        self.process_files(live_post=live_post.value, files=files)

        live_page.update_live_post(live_post=live_post)

    def delete_message(self, message):
        """
        Deletes a message in the live page corresponding to the channel where
        the message was posted if such a page exists.

        Args:
            message (dict): A message deleted from a messaging app.
        """

        channel_id = self.get_channel_id_from_message(message=message)
        try:
            live_page = self.get_live_page_from_channel_id(channel_id=channel_id)
        except self.model.DoesNotExist:
            return

        message_id = self.get_message_id_from_edited_message(message=message)
        live_page.delete_live_post(message_id=message_id)


@method_decorator(csrf_exempt, name="dispatch")
class WebhookReceiverMixin(View):
    """
    Mixin for receivers using the webhook technique.

    Attributes:
        url_path (str):
            Path of the URL used by a messaging app to send new updates to this receiver.
        url_name (str):
            Name of the URL for reversing/resolving.
    """

    url_path = ""
    url_name = ""

    def verify_request(self, request, body, *args, **kwargs):
        """
        Ensures that the incoming request comes from the messaging app expected.

        Args:
            request (HttpRequest): Http request
            body (str): Body of the request

        Raises:
            RequestVerificationError: if the request verification failed
        """

        raise NotImplementedError

    def post(self, request, *args, **kwargs):
        """
        This is the main method for Webhook receivers.

        It handles new updates from messaging apps in these three steps:

        1. Verify the request.
        2. Dispatch the new event and process the updates received.
        3. Acknowledge the request.

        Args:
            request (HttpRequest): Http request

        Returns:
            HttpResponse:
            - `Forbidden` if the request couldn't be verified.
            - `OK` if the request is verified and updates have been succesfully processed.
        """

        body = request.body.decode("utf-8")
        try:
            self.verify_request(request, body, *args, **kwargs)
        except RequestVerificationError:
            return HttpResponseForbidden("Request verification failed.")

        self.dispatch_event(event=json.loads(body))
        return HttpResponse("OK")

    @classmethod
    def webhook_connection_set(cls):
        """
        Checks if webhook a connection is set.

        We call this method before calling the `set_webhook` method in order to
        avoid sending unneccesary POST requests to the messaging app server.

        Returns:
            bool:
            - `True` if a webhook connection is set,
            - `False` else.
        """

        raise NotImplementedError

    @classmethod
    def set_webhook(cls):
        """
        Sets a webhook connection with the corresponding messaging app.

        This method may be trivial for messaging apps which propose setting a webhook
        in their UI like Slack. It may also be the main method if we have to set up
        the webhook connection ourselves; like with Telegram for example.

        Raises:
            WebhookSetupError: if the webhook connection with the messaging app failed.
        """

        raise NotImplementedError

    @classmethod
    def get_urls(cls):
        """
        Retrieves webhook urls after having ensured that a webhook connection is enabled with
        the corresponding messaging app.

        Returns:
            URLPattern: URL which messaging apps use to send new updates.

        Raises:
            WebhookSetupError: if the webhook connection with the messaging app didn't succeed.
        """

        if not cls.webhook_connection_set():
            cls.set_webhook()

        return [
            path(cls.url_path, cls.as_view(), name=cls.url_name),
        ]



""" Webapp models."""

from django.db import models


class Channel(models.Model):
    """A model that mimics a channel in a messaging app."""

    channel_name = models.CharField(
        max_length=124,
        help_text="Channel name",
        primary_key=True,
    )
    created = models.DateTimeField(
        auto_now_add=True,
        help_text="Date and time of channel creation",
    )

    def __str__(self):
        return f"#{self.channel_name}"

    class Meta:
        ordering = ["-created"]


class Message(models.Model):
    """A model that mimics a message in a messaging app."""

    channel = models.ForeignKey(
        "Channel",
        on_delete=models.CASCADE,
        related_name="messages",
        help_text="Channel this message was posted to.",
    )
    created = models.DateTimeField(
        auto_now_add=True,
        help_text="Date and time of message creation",
    )
    modified = models.DateTimeField(
        null=True,
        blank=True,
        help_text="Date and time of last update",
    )
    show = models.BooleanField(
        default=True,
        blank=True,
        help_text="Indicates if this message is shown/hidden",
    )
    content = models.TextField(help_text="Content of the message")

    @property
    def images(self):
        return self.files.all()

    class Meta:
        ordering = ["-created"]


class Image(models.Model):
    """A Model that represents an image of a message."""

    image = models.ImageField(help_text="Image of a message")
    message = models.ForeignKey(
        "Message",
        on_delete=models.CASCADE,
        related_name="files",
        help_text="Message this image was posted to.",
    )



import json

import aioredis
from asgiref.sync import async_to_sync

from ..utils import get_redis_url
from ..websocket import BaseWebsocketPublisher


def make_channel_group_name(channel_id):
    return f"group_{channel_id}"


async def redis_publish(channel_group_name, message):
    """
    Publishes a message to the given channel in Redis.

    Args:
        channel_group_name (str):
            Channel to publish the message to.
        message (*):
            Message to publish.
    """

    redis = aioredis.from_url(get_redis_url())
    await redis.publish(channel_group_name, json.dumps(message))


class RedisPubSubPublisher(BaseWebsocketPublisher):
    """Publisher using Redis PubSub functionnality."""

    def publish(self, channel_id, renders, removals):
        """
        Publishes in Redis the renders and removals to the channel group
        corresponding to channel_id.

        See base class.
        """

        channel_group_name = make_channel_group_name(channel_id)
        message = {"renders": renders, "removals": removals}

        async_to_sync(redis_publish)(channel_group_name, message)



from django.apps import AppConfig


class WagtailLiveConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "wagtail_live"

    def ready(self):
        from wagtail_live.publishers.websocket import BaseWebsocketPublisher
        from wagtail_live.utils import get_live_publisher

        live_publisher = get_live_publisher()

        # Connect a listener to the live_page_update signal
        # if the publisher defined uses the websockets technique
        if issubclass(live_publisher, BaseWebsocketPublisher):
            from wagtail_live.signals import live_page_update

            # Set`weak=False` to avoid the publisher being garbage collected.
            # See:
            # https://docs.djangoproject.com/en/3.2/topics/signals/#django.dispatch.Signal.connect
            live_page_update.connect(
                live_publisher(), weak=False, dispatch_uid="live_publisher"
            )



from django import template

from wagtail_live.publishers.piesocket import (
    get_piesocket_api_key,
    get_piesocket_endpoint,
)
from wagtail_live.publishers.utils import get_live_server_host, get_live_server_port

register = template.Library()


@register.simple_tag
def piesocket_api_key():
    return get_piesocket_api_key()


@register.simple_tag
def piesocket_endpoint():
    return get_piesocket_endpoint()


@register.simple_tag
def get_server_host():
    return get_live_server_host()


@register.simple_tag
def get_server_port():
    return get_live_server_port()



from .receivers import TelegramWebhookMixin, TelegramWebhookReceiver
from .utils import (
    get_base_telegram_url,
    get_telegram_bot_token,
    get_telegram_webhook_url,
)

__all__ = [
    "TelegramWebhookMixin",
    "TelegramWebhookReceiver",
    "get_base_telegram_url",
    "get_telegram_bot_token",
    "get_telegram_webhook_url",
]



from django.core.management.base import BaseCommand, CommandError


class Command(BaseCommand):
    help = "Run a separate server for Wagtail Live publishers."

    def add_arguments(self, parser):
        parser.add_argument("publisher", nargs="+", type=str)

    def handle(self, *args, **options):
        publisher = options["publisher"][0]
        if publisher == "websockets":
            import asyncio

            from wagtail_live.publishers.websockets import app

            asyncio.run(app())

        elif publisher == "starlette":
            import uvicorn

            from wagtail_live.publishers.starlette import app
            from wagtail_live.publishers.utils import (
                get_live_server_host,
                get_live_server_port,
            )

            host, port = get_live_server_host(), get_live_server_port()

            uvicorn.run(app, host=host, port=port)

        else:
            raise CommandError('Publisher "%s" does not exist' % publisher)



from .version import get_version

# major.minor.patch.release.number
# release must be one of alpha, beta, rc, or final
VERSION = (1, 0, 0, "final", 1)

__version__ = get_version(VERSION)

# For compatibility with Django < 3.2
default_app_config = "wagtail_live.apps.WagtailLiveConfig"



from wagtail_live.receivers.base import BaseMessageReceiver

from .models import Image

MESSAGE_CREATED = 1
MESSAGE_EDITED = 2
MESSAGE_DELETED = 3


class WebAppReceiver(BaseMessageReceiver):
    """Webapp receiver."""

    def dispatch_event(self, event):
        """See base class."""

        message = event
        if message["update_type"] == MESSAGE_EDITED:
            self.change_message(message=message)
            return

        elif message["update_type"] == MESSAGE_DELETED:
            self.delete_message(message=message)
            return

        else:
            self.add_message(message=message)

    def get_channel_id_from_message(self, message):
        """See base class."""

        return message["channel"]

    def get_message_id_from_message(self, message):
        """See base class."""

        return message["id"]

    def get_message_text(self, message):
        """See base class."""

        return message["content"]

    def get_message_files(self, message):
        """See base class."""

        return message["images"]

    def get_image_title(self, image):
        """See base class."""

        return image["image"]["name"].split(".")[0].replace("-", " ")

    def get_image_name(self, image):
        """See base class."""

        return image["image"]["name"]

    def get_image_mimetype(self, image):
        """See base class."""

        mime_type = image["image"]["name"].split(".")[-1]
        return "jpeg" if mime_type == "jpg" else mime_type

    def get_image_dimensions(self, image):
        """See base class."""

        return (image["image"]["width"], image["image"]["height"])

    def get_image_content(self, image):
        """See base class."""

        return Image.objects.get(id=image["id"]).image

    def get_message_id_from_edited_message(self, message):
        """See base class."""

        return self.get_message_id_from_message(message=message)

    def get_message_text_from_edited_message(self, message):
        """See base class."""

        return self.get_message_text(message=message)

    def get_message_files_from_edited_message(self, message):
        """See base class."""

        return self.get_message_files(message=message)



import asyncio
import json

from starlette.applications import Starlette
from starlette.endpoints import WebSocketEndpoint
from starlette.middleware.cors import CORSMiddleware

from ..redis import RedisBus, make_channel_group_name
from ..utils import get_redis_url


# Define the broadcast method to be used by the bus.
async def broadcast(message, connections):
    message = json.loads(message)
    await asyncio.wait([ws.send_json(message) for ws in connections])


BUS = RedisBus(url=get_redis_url(), broadcast=broadcast)


app = Starlette()

# TODO: Forgot why I did this.
app.add_middleware(CORSMiddleware, allow_origins=["*"])


@app.on_event("startup")
def startup():
    # Run the bus on startup
    asyncio.create_task(BUS.run())


@app.websocket_route("/ws/channel/{channel_id:str}/")
class StarlettePublisherApp(WebSocketEndpoint):
    encoding = "json"

    async def on_connect(self, websocket, **kwargs):
        """Adds this connection to the channel group corresponding to channel_id."""

        await websocket.accept()

        channel_id = websocket.path_params["channel_id"]
        self.channel_group_name = make_channel_group_name(channel_id)
        await BUS.subscribe(self.channel_group_name, websocket)

    async def on_disconnect(self, websocket, close_code):
        """Removes this connection from its channel group."""

        await BUS.unsubscribe(self.channel_group_name, websocket)



import asyncio
from collections import defaultdict

import aioredis

from ..utils import get_redis_url


class RedisBus:
    """
    Class providing an event bus based on Redis PubSub.

    A channel group is created for each live page.

    Users viewing the page in live are added to that channel group.

    When a live page is updated, a message is sent to its corresponding
    channel group. The message sent contains the renders and the removals
    i.e the new updates of the page.

    The message is then broadcasted to users that have subscribed to that
    channel group.

    Attributes:
        pubsub (PubSub):
            Redis PubSub class which allows to do pub/sub operations.
        broadcast (callable):
            The function to use when broadcasting a message to clients.
        channel_groups (dict):
            Maps a channel group to the connections that have subscribed to it.

    """

    def __init__(self, url, broadcast):
        redis = aioredis.from_url(get_redis_url(), decode_responses=True)
        self.pubsub = redis.pubsub(ignore_subscribe_messages=True)
        self.broadcast = broadcast
        self.channel_groups = defaultdict(set)
        # Tracks when the bus starts running i.e on the first connection.
        self._running = None

    async def run(self):
        """Retrieves messages from Redis and dispatches them as long as the server is running."""

        # We will have an error if we try to run the bus if we haven't
        # subscribed to any channel yet.
        # Therefore, wait for the bus to start running before calling pubsub.run.
        self._running = asyncio.Event()
        await self._running.wait()
        await self.pubsub.run()

    def _set_running(self):
        if self._running is not None and not self._running.is_set():
            self._running.set()

    def get_channel_group_subscribers(self, channel_group_name):
        """
        Retrieves the connections that have subscribed to channel_group_name.

        Args:
            channel_group_name (str):
                The channel group to find subscribers for.

        Returns:
            Set: Connections/users that have subscribed to the given channel group.
        """

        return self.channel_groups[channel_group_name]

    def handle_message(self, message):
        """
        Callback called when a message is published on a channel.

        Retrieves the connections that should receive the message and
        spawns a task to broadcast them the message.

        Args:
            message (str): The message published on Redis.
        """

        connections = self.get_channel_group_subscribers(message["channel"])
        asyncio.create_task(self.broadcast(message["data"], connections))

    async def subscribe(self, channel_group_name, ws_connection):
        """
        Subscribes a connection to a channel group.

        Args:
            channel_group_name (str):
                Channel group to subscribe to
            ws_connection (*):
                The websocket or connection instance to add to the channel_group subscribers.
                It must have a method to send messages.
        """

        # Subscribe to this channel group in Redis if not done yet.
        if not self.get_channel_group_subscribers(channel_group_name):
            # Passing parameters as `channel=handler` to the pubsub.subscribe method
            # has the effect to call `handler` whenever a message is published on `channel`.
            await self.pubsub.subscribe(**{channel_group_name: self.handle_message})

        self.channel_groups[channel_group_name].add(ws_connection)
        self._set_running()

    async def unsubscribe(self, channel_group_name, ws_connection):
        """
        Unsubscribes a connection from a channel group.

        Args:
            channel_group_name (str):
                Channel group to unsubscribe from.
            ws_connection (*):
                The websocket or connection instance to remove from the channel_group subscribers.

        """

        self.channel_groups[channel_group_name].remove(ws_connection)

        # Unsubscribe from this channel_group in Redis if no more
        # connections are subscribed to it.
        if not self.get_channel_group_subscribers(channel_group_name):
            await self.pubsub.unsubscribe(channel_group_name)
            del self.channel_groups[channel_group_name]



"""Wagtail Live Exceptions."""


class RequestVerificationError(Exception):
    """A incoming request couldn't be verified."""

    pass


class WebhookSetupError(Exception):
    """An error occured while setting a webhook connection with the input source."""

    pass



""" Wagtail Live models."""

from django.conf import settings
from django.db import models
from django.utils import timezone
from wagtail.admin.edit_handlers import FieldPanel, StreamFieldPanel
from wagtail.core.fields import StreamField

from wagtail_live.blocks import LivePostBlock, compare_live_posts_values
from wagtail_live.signals import live_page_update


class LivePageMixin(models.Model):
    """
    Base class for pages using Wagtail Live.

    Attributes:
        channel_id (str):
            Id of the corresponding channel in a messaging app.
        last_updated_at (DateTime):
            Date and time of the last update of this page.
        live_posts (StreamField):
            StreamField containing all the posts/messages published
            respectively on this page/channel.
    """

    channel_id = models.CharField(
        help_text="Channel ID",
        max_length=255,
        blank=True,
        unique=True,
    )
    last_updated_at = models.DateTimeField(
        help_text="Last update of this page",
        blank=True,
        default=timezone.now,
    )

    live_posts = StreamField(
        [
            ("live_post", LivePostBlock()),
        ],
        blank=True,
    )

    panels = [
        FieldPanel("channel_id"),
        StreamFieldPanel("live_posts"),
    ]

    @property
    def last_update_timestamp(self):
        """Timestamp of the last update of this page."""

        # Live posts are saved using a json format.
        # We strip the microseconds here to follow that format.
        microsecond = (self.last_updated_at.microsecond // 1000) * 1000
        return self.last_updated_at.replace(microsecond=microsecond).timestamp()

    def save(self, sync=True, *args, **kwargs):
        """Update live page on save depending on the `WAGTAIL_LIVE_SYNC_WITH_ADMIN` setting."""

        sync_changes = sync and getattr(settings, "WAGTAIL_LIVE_SYNC_WITH_ADMIN", True)
        has_changed = False
        if sync_changes and self.id:
            renders, seen = [], set()
            previous_posts = {
                live_post.id: live_post
                for live_post in self.__class__.objects.get(id=self.id).live_posts
            }
            now = timezone.now()

            for i, post in enumerate(self.live_posts):  # New posts
                post_id = post.id
                if post_id in previous_posts:
                    seen.add(post_id)

                    # Check if the post has been modified.
                    previous_post = previous_posts[post_id]
                    identic = compare_live_posts_values(post.value, previous_post.value)
                    if not identic:
                        post.value["modified"] = now
                        renders.append(i)

                else:
                    # This is a new post.
                    # Force the value of `created` here to keep it synchronized with the
                    # `last_updated_at` property.
                    # This is mostly to avoid missing new updates with the polling publishers.
                    post.value["created"] = now
                    renders.append(i)

            removals = list(set(previous_posts.keys()).difference(seen))

            has_changed = bool(renders or removals)
            if has_changed:
                self.last_updated_at = now

        result = super().save(*args, **kwargs)

        if sync_changes and has_changed:
            # Reverse renders so the latest posts, which are in the start of the list,
            # are processed later in the front end.
            renders.reverse()
            renders = list(map(self.get_live_post_by_index, renders))

            # Send signal.
            live_page_update.send(
                sender=self.__class__,
                channel_id=self.channel_id,
                renders=renders,
                removals=removals,
            )

        return result

    def _get_live_post_index(self, message_id):
        for i, post in enumerate(self.live_posts):
            if post.value["message_id"] == message_id:
                return i

    def get_live_post_index(self, message_id):
        """
        Retrieves the index of a live post.

        Args:
            message_id (str): ID of the message corresponding to a live post.

        Returns:
            int: Index of the live post if found else -1
        """

        return self._get_live_post_index(message_id=message_id)

    def get_live_post_by_index(self, live_post_index):
        """
        Retrieves a live post by its index.

        Args:
            live_post_index (int): Index of the live post to look for.

        Returns:
            LivePostBlock: The live post instance

        Raises:
            IndexError: if a live post with the given index doesn't exist.
        """

        return self.live_posts[live_post_index]

    def get_live_post_by_message_id(self, message_id):
        """
        Retrieves a live post by its ID.

        Args:
            message_id (str):
                ID of the message corresponding to a live post.

        Returns:
            LivePostBlock: The live post instance

        Raises:
            KeyError: if a live post with the given ID doesn't exist.
        """

        live_post_index = self.get_live_post_index(message_id=message_id)
        if live_post_index is None:
            raise KeyError
        return self.get_live_post_by_index(live_post_index)

    def add_live_post(self, live_post):
        """
        Adds a new live post to live page.

        Args:
            live_post (LivePostBlock):
                live post to add
        """

        posts = self.live_posts
        lp_index = 0
        post_created_at = live_post["created"]
        while lp_index < len(posts):
            if posts[lp_index].value["created"] < post_created_at:
                break
            lp_index += 1

        # Insert to keep posts sorted by time
        self.live_posts.insert(lp_index, ("live_post", live_post))

        self.last_updated_at = post_created_at
        self.save(sync=False)

        live_post = self.get_live_post_by_index(lp_index)
        live_page_update.send(
            sender=self.__class__,
            channel_id=self.channel_id,
            renders=[live_post],
            removals=[],
        )

    def update_live_post(self, live_post):
        """
        Updates a live post when it has been edited.

        Args:
            live_post (livePostBlock):
                Live post to update.
        """

        live_post.value["modified"] = self.last_updated_at = timezone.now()
        self.save(sync=False)

        live_page_update.send(
            sender=self.__class__,
            channel_id=self.channel_id,
            renders=[live_post],
            removals=[],
        )

    def delete_live_post(self, message_id):
        """
        Deletes the live post corresponding to message_id.

        Args:
            message_id (str):
                ID of the message corresponding to a live post.
        Raises:
            KeyError: if live post containing message with message_id doesn't exist.
        """

        live_post_index = self.get_live_post_index(message_id=message_id)
        if live_post_index is None:
            raise KeyError

        live_post_id = self.live_posts[live_post_index].id
        del self.live_posts[live_post_index]

        self.last_updated_at = timezone.now()
        self.save(sync=False)

        live_page_update.send(
            sender=self.__class__,
            channel_id=self.channel_id,
            renders={},
            removals=[live_post_id],
        )

    def get_updates_since(self, last_update_ts):
        """
        Retrieves new updates since a given timestamp value.

        Args:
            last_update_ts (DateTime):
                Timestamp of the last update.

        Returns:
            (list, dict):
                a tuple containing the current live posts
                and the updated posts since `last_update_ts`.
        """

        # Reverse posts list so that latest updates are processed later by the client side.
        posts = reversed(self.live_posts)
        current_posts, updated_posts = [], {}
        for post in posts:
            post_id = post.id
            current_posts.append(post_id)

            created = post.value["created"]
            if created >= last_update_ts:  # This is a new post
                updated_posts[post_id] = {
                    "show": post.value["show"],
                    "content": post.render(context={"block_id": post.id}),
                }
                continue

            last_modified = post.value["modified"]
            if last_modified and last_modified >= last_update_ts:
                # This is an edited post
                updated_posts[post_id] = {
                    "show": post.value["show"],
                    "content": post.render(context={"block_id": post.id}),
                }

        return (updated_posts, current_posts)

    class Meta:
        abstract = True



"""Webapp serializers."""

from rest_framework import serializers

from .models import Channel, Image, Message


class ImageListSerializer(serializers.ListSerializer):
    def get_value(self, dictionary):
        if hasattr(dictionary, "getlist"):
            images = dictionary.getlist("images")
        else:
            images = dictionary.get("images")

        if not images or not images[0]:
            return
        return [{"image": item} for item in images]


class ImageSerializer(serializers.ModelSerializer):
    """Image serializer"""

    class Meta:
        model = Image
        fields = ["image"]
        list_serializer_class = ImageListSerializer

    def to_representation(self, instance):
        image = instance.image
        return {
            "id": instance.id,
            "image": {
                "name": image.name,
                "url": image.url,
                "width": image.width,
                "height": image.height,
            },
        }


class MessageSerializer(serializers.ModelSerializer):
    """Message serializer"""

    images = ImageSerializer(many=True, allow_null=True)

    class Meta:
        model = Message
        fields = ["id", "channel", "created", "modified", "show", "content", "images"]

    def create(self, validated_data):
        images = validated_data.pop("images")
        message = Message.objects.create(**validated_data)

        if images:
            for image in images:
                Image.objects.create(message=message, **image)

        return message

    def update(self, instance, validated_data):
        images = validated_data.pop("images")
        for attr, value in validated_data.items():
            setattr(instance, attr, value)

        if images:
            for image in images:
                Image.objects.create(message=instance, **image)

        instance.save()
        return instance


class ChannelSerializer(serializers.ModelSerializer):
    """Channel serializer"""

    class Meta:
        model = Channel
        fields = ["channel_name", "created"]



import hmac
import json
import re
import time
from hashlib import sha256

import requests
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.core.files.base import ContentFile
from django.core.validators import URLValidator
from django.http import HttpResponse

from wagtail_live.exceptions import RequestVerificationError
from wagtail_live.receivers.base import BaseMessageReceiver, WebhookReceiverMixin
from wagtail_live.utils import is_embed


class SlackWebhookMixin(WebhookReceiverMixin):
    """Slack WebhookMixin."""

    url_path = "slack/events"
    url_name = "slack_events_handler"

    def post(self, request, *args, **kwargs):
        """
        Checks if Slack is trying to verify our Request URL.

        Returns:
            HttpResponse:
                containing the challenge string if Slack is trying to verify our request URL.
        """

        payload = json.loads(request.body.decode("utf-8"))
        if payload["type"] == "url_verification":
            return HttpResponse(payload["challenge"])
        return super().post(request, *args, **kwargs)

    @staticmethod
    def sign_slack_request(content):
        """Signs content from a Slack request using the `SLACK_SIGNING_SECRET` as key."""

        hasher = hmac.new(str.encode(settings.SLACK_SIGNING_SECRET), digestmod=sha256)
        hasher.update(str.encode(content))
        return hasher.hexdigest()

    def verify_request(self, request, body):
        """
        Verifies Slack requests.
        See: [Verifying requests from Slack](https://api.slack.com/authentication/
        verifying-requests-from-slack).

        Args:
            request (HttpRequest): from Slack

        Raises:
            RequestVerificationError: if request failed to be verified.
        """

        timestamp = request.headers.get("X-Slack-Request-Timestamp")
        if not timestamp:
            raise RequestVerificationError(
                "X-Slack-Request-Timestamp not found in request's headers."
            )

        if abs(time.time() - float(timestamp)) > 60 * 5:
            # The request timestamp is more than five minutes from local time.
            # It could be a replay attack, so let's ignore it.
            raise RequestVerificationError(
                "The request timestamp is more than five minutes from local time."
            )

        sig_basestring = "v0:" + timestamp + ":" + body
        my_signature = "v0=" + self.sign_slack_request(content=sig_basestring)
        slack_signature = request.headers["X-Slack-Signature"]
        if not hmac.compare_digest(slack_signature, my_signature):
            raise RequestVerificationError("Slack signature couldn't be verified.")

    @classmethod
    def set_webhook(cls):
        """This is done in Slack UI."""

        pass

    @classmethod
    def webhook_connection_set(cls):
        """Assume that it's true."""

        return True


class SlackEventsAPIReceiver(BaseMessageReceiver, SlackWebhookMixin):
    """Slack Events API receiver."""

    def dispatch_event(self, event):
        """See base class."""

        message = event["event"]

        subtype = message.get("subtype")
        if subtype:
            if subtype == "message_changed":
                self.change_message(message=message)
            elif subtype == "message_deleted":
                self.delete_message(message=message)
            elif subtype == "file_share":
                self.add_message(message=message)
            return

        self.add_message(message=message)

    def get_channel_id_from_message(self, message):
        """See base class."""

        return message["channel"]

    def get_message_id_from_message(self, message):
        """See base class."""

        return message["ts"]

    def get_message_text(self, message):
        """See base class."""

        return message["text"]

    def get_message_files(self, message):
        """See base class."""

        return message["files"] if "files" in message else []

    def get_image_title(self, image):
        """See base class."""

        return image["title"]

    def get_image_name(self, image):
        """See base class."""

        return image["name"]

    def get_image_mimetype(self, image):
        """See base class."""

        return image["mimetype"].split("/")[1]

    def get_image_dimensions(self, image):
        """See base class."""

        try:
            return (image["original_w"], image["original_h"])
        except KeyError:
            raise ValueError

    def get_image_content(self, image):
        """See base class."""

        slack_bot_token = getattr(settings, "SLACK_BOT_TOKEN", "")
        if not slack_bot_token:
            raise ImproperlyConfigured(
                "You haven't specified SLACK_BOT_TOKEN in your settings."
                + "You won't be able to upload images from Slack without this setting defined."
            )
        headers = {"Authorization": f"Bearer {slack_bot_token}"}
        response = requests.get(image["url_private"], headers=headers)
        return ContentFile(response.content)

    def get_message_id_from_edited_message(self, message):
        """See base class."""

        return self.get_message_id_from_message(message=message["previous_message"])

    def get_message_text_from_edited_message(self, message):
        """See base class."""

        return self.get_message_text(message=message["message"])

    def get_message_files_from_edited_message(self, message):
        """See base class."""

        return self.get_message_files(message=message["message"])

    def get_embed(self, text):
        """
        Slack sends url in this format:
        `<https://twitter.com/wagtail/|https://twitter.com/wagtail/>`
        where the first part is the full url and the second part represents the user's input.

        See [Links in retrieved messages](https://api.slack.com/reference/surfaces/
        formatting#links-in-retrieved-messages).
        """

        # Check if the text provided is a Slack-like url
        if text.startswith("<") and text.endswith(">"):
            # Get the url resolved by Slack
            url = text[1:-1].split("|")[0]
            if is_embed(text=url):
                return url

        return ""

    def parse_text(self, text):
        """
        See [Links in retrieved messages](https://api.slack.com/reference/surfaces/
        formatting#links-in-retrieved-messages).

        See also base class.
        """

        url_format = re.compile(r"<http([^|]+?)(\|([^|]+?))?>")

        urls = url_format.finditer(text)
        for url_match in urls:
            match = url_match.group()[1:-1]
            if "|" in match:
                url, description = match.split("|")
            else:
                url = description = match

            try:
                validator = URLValidator()
                validator(url)
            except ValidationError:
                continue

            text = text.replace(url_match.group(), f"<a href='{url}'>{description}</a>")

        return text



import time
from datetime import datetime, timezone

from django.conf import settings
from django.http import JsonResponse
from django.shortcuts import get_object_or_404
from django.urls import path
from django.utils.functional import cached_property
from django.views import View

from wagtail_live.utils import (
    get_live_page_model,
    get_polling_interval,
    get_polling_timeout,
)


class PollingPublisherMixin(View):
    """
    A mixin for publishers using the polling technique.

    Attributes:
        url_path (str):
            Path of the URL used by client side to fetch new updates.
        url_name (str):
            Name of the URL for reversing/resolving.
    """

    url_path = "get-updates/<str:channel_id>/"
    url_name = ""

    @cached_property
    def model(self):
        """Retrieves the live model defined."""

        return get_live_page_model()

    @classmethod
    def get_urls(cls):
        """Retrieves the URLs client side uses to fetch updates."""

        return [
            path(cls.url_path, cls.as_view(), name=cls.url_name),
        ]

    @staticmethod
    def get_last_update_client_from_request(request):
        """
        Retrieves the timestamp of the last update received in the client side.

        Args:
            request (HttpRequest): client side request

        Returns:
            float: Timestamp of the last update received in the client side.
        """

        return float(request.GET.get("last_update_ts"))

    def post(self, request, channel_id, *args, **kwargs):
        """
        Initiates communication with client side and sends current live posts.

        Args:
            request (HttpRequest):
                Client side's request
            channel_id (str):
                Id of the channel to get updates from.

        Returns:
            HttpResponse:
            - JSONResponse with the following informations:
                - A list of the IDs of the current live posts for the page requested.

                    Client side uses this list to keep track of live posts that have been deleted.

                - Timestamp of the last update for the page requested.

                    Client side uses this to know when new updates are available.

                - The duration of the polling interval for interval polling.

                if a page corresponding to the `channel_id` given exists.

            - Http404 else.
        """

        raise NotImplementedError

    def get(self, request, channel_id, *args, **kwargs):
        """
        Retrieves and sends new updates to the client side.

        Args:
            request (HttpRequest):
                Client side's request sent along with the timestamp of the
                last update received.
            channel_id (str):
                Id of the channel to get last update's timestamp from.

        Returns:
            HttpResponse:
            - JSONResponse with the following informations:
                - A mapping of the live posts updated since client side's last update timestamp.

                    Keys represents IDs of the live posts edited and the values
                    are the new content of those live posts.

                - A list of the IDs of the current live posts for the page requested.

                    Client side compares this list to the one it has and remove the live posts
                    whose IDs aren't in this new list.

                - Timestamp of the last update for the page requested.

                if a page corresponding to the `channel_id` given exists.

            - Http404 else.
        """

        raise NotImplementedError


class IntervalPollingPublisher(PollingPublisherMixin):
    """
    Interval polling Publisher. Class Based View.

    This class handles delivering new updates to the client side (Interval polling technique).
    It accepts 3 request methods (POST, HEAD and GET) which correspond to the following steps:

    1. The client side initiates (shake) the communication by sending a POST request.
        The publisher acknowledges by sending relevant data to the client side.

    2. The client side asks if there are any updates by sending a HEAD request.
        If no updates are available, client side sleeps for the duration of the polling interval
        and repeats this step.

    3. If new updates are available, client side sends a GET request to get the new updates.
    """

    url_name = "interval-polling"

    def post(self, request, channel_id, *args, **kwargs):
        """See base class."""

        live_page = get_object_or_404(self.model, channel_id=channel_id)
        return JsonResponse(
            {
                "livePosts": [live_post.id for live_post in live_page.live_posts],
                "lastUpdateTimestamp": live_page.last_update_timestamp,
                "pollingInterval": get_polling_interval(),
            }
        )

    def head(self, request, channel_id, *args, **kwargs):
        """
        Sends the timestamp of the last update for the page requested.

        Args:
            request (HttpRequest):
                Client side's request
            channel_id (str):
                Id of the channel to get last update's timestamp from.

        Returns:
            HttpResponse:
            - OK: if a page corresponding to the `channel_id` given exists.

                The timestamp of the last update for the page requested is sent in the response.
                Client side checks if this value is greater than the last one received.
                In such case, the client side knows that new updates are available and sends a
                GET request to get those updates.

            - Http404: else.
        """

        live_page = get_object_or_404(self.model, channel_id=channel_id)
        response = JsonResponse(data={}, status=200)
        response["Last-Update-At"] = live_page.last_update_timestamp
        return response

    def get(self, request, channel_id, *args, **kwargs):
        """See base class."""

        live_page = get_object_or_404(self.model, channel_id=channel_id)
        last_update_client = self.get_last_update_client_from_request(request=request)
        tz = timezone.utc if settings.USE_TZ else None
        updated_posts, current_posts = live_page.get_updates_since(
            last_update_ts=datetime.fromtimestamp(last_update_client, tz=tz),
        )

        return JsonResponse(
            {
                "updates": updated_posts,
                "currentPosts": current_posts,
                "lastUpdateTimestamp": live_page.last_update_timestamp,
            }
        )


class LongPollingPublisher(PollingPublisherMixin):
    """
    Long polling Publisher. Class Based View.

    This class handles delivering new updates to the client side (Long polling technique).
    It accepts 2 request methods (POST and GET) which correspond to the following steps:

    1. The client side initiates (shake) the communication by sending a POST request.
        The publisher acknowledges by sending relevant data to the client side.

    2. The client side asks new updates by sending a GET request.

        Server side keeps the connection open until a new update is available.
        In that case, updates are directly sent to client side.

        If the polling timeout duration is reached, it sends a response
        containing a `timeOutReached` parameter which indicates the client side
        that there aren't updates available.
    """

    url_name = "long-polling"

    def post(self, request, channel_id, *args, **kwargs):
        """See base class."""

        live_page = get_object_or_404(self.model, channel_id=channel_id)
        return JsonResponse(
            {
                "livePosts": [live_post.id for live_post in live_page.live_posts],
                "lastUpdateTimestamp": live_page.last_update_timestamp,
            }
        )

    def get(self, request, channel_id, *args, **kwargs):
        """
        Sends updates when they are available as long as the polling timeout isn't reached.

        See base class.
        """

        last_update_client = self.get_last_update_client_from_request(request=request)
        polling_timeout = get_polling_timeout()
        starting_time = time.time()

        while time.time() - starting_time < polling_timeout:
            live_page = get_object_or_404(self.model, channel_id=channel_id)
            last_update_ts = live_page.last_update_timestamp
            if last_update_ts > last_update_client:
                tz = timezone.utc if settings.USE_TZ else None
                updated_posts, current_posts = live_page.get_updates_since(
                    last_update_ts=datetime.fromtimestamp(last_update_client, tz=tz),
                )

                return JsonResponse(
                    {
                        "updates": updated_posts,
                        "currentPosts": current_posts,
                        "lastUpdateTimestamp": last_update_ts,
                    }
                )

            # Maybe propose a setting so the user can define this value
            time.sleep(0.5)

        return JsonResponse({"timeOutReached": "Timeout duration reached."})



import logging

import requests
from django.core.files.base import ContentFile

from wagtail_live.exceptions import RequestVerificationError, WebhookSetupError
from wagtail_live.receivers.base import BaseMessageReceiver, WebhookReceiverMixin
from wagtail_live.utils import is_embed

from .utils import (
    format_url,
    get_base_telegram_url,
    get_telegram_bot_token,
    get_telegram_webhook_url,
)

logger = logging.getLogger(__name__)


class TelegramWebhookMixin(WebhookReceiverMixin):
    """Telegram WebhookMixin."""

    url_path = "telegram/events/<str:token>/"
    url_name = "telegram_events_handler"

    def verify_request(self, request, body, token, *args, **kwargs):
        """See base class."""

        if token != get_telegram_bot_token():
            raise RequestVerificationError

    @classmethod
    def webhook_connection_set(cls):
        """See base class."""

        response = requests.get(get_base_telegram_url() + "getWebhookInfo")
        if response.ok:
            # Ensure that the webhook is set with the correct URL
            payload = response.json()
            return (
                payload["ok"] and payload["result"]["url"] == get_telegram_webhook_url()
            )
        return False

    @classmethod
    def set_webhook(cls):
        """Sets webhook connection with Telegram's API"""

        response = requests.get(
            get_base_telegram_url() + "setWebhook",
            params={
                "url": get_telegram_webhook_url(),
                "allowed_updates": [
                    "message",
                    "edited_message",
                    "channel_post",
                    "edited_channel_post",
                ],
            },
        )
        payload = response.json()

        if not response.ok or not payload["ok"]:
            raise WebhookSetupError(
                "Failed to set Webhook connection with Telegram's API. "
                + f"{payload['description']}"
            )


class TelegramWebhookReceiver(TelegramWebhookMixin, BaseMessageReceiver):
    """Telegram webhook receiver."""

    def dispatch_event(self, event):
        """
        **Note**: Telegram doesn't send an update when a message is deleted.

        See also base class.
        """

        for edit_type in ["edited_message", "edited_channel_post"]:
            if edit_type in event:
                self.change_message(message=event.get(edit_type))
                return

        message = event.get("message") or event.get("channel_post")
        if "media_group_id" in message:
            try:
                self.add_image_to_message(message=message)
                return
            except KeyError:
                pass

        if "entities" in message and message["entities"][0]["type"] == "bot_command":
            self.handle_bot_command(message=message)
            return

        self.add_message(message=message)

    def add_image_to_message(self, message):
        """Telegram sends images uploaded in a message one by one."""

        channel_id = self.get_channel_id_from_message(message=message)
        try:
            live_page = self.get_live_page_from_channel_id(channel_id=channel_id)
        except self.model.DoesNotExist:
            return

        message_id = self.get_message_id_from_message(message=message)
        live_post = live_page.get_live_post_by_message_id(message_id=message_id)

        files = self.get_message_files(message=message)
        self.process_files(live_post=live_post.value, files=files)

        live_page.update_live_post(live_post=live_post)

    def handle_bot_command(self, message):
        """
        Handles the following bot commands:

        - `/get_chat_id`: returns the id of the current chat.
        """

        command = message["entities"][0]
        start = command["offset"]
        end = start + command["length"]
        command_text = message["text"][start:end]

        if command_text == "/get_chat_id":
            chat_id = self.get_channel_id_from_message(message=message)
            response = requests.get(
                get_base_telegram_url() + "sendMessage",
                params={
                    "chat_id": chat_id,
                    "text": chat_id,
                },
            )

            payload = response.json()
            if not payload["ok"]:
                logger.error(payload["description"])

    def get_channel_id_from_message(self, message):
        """See base class."""

        # Since live posts aren't cleaned when they are added via a receiver,
        # we make sure at this level that we return the correct types.
        return str(message["chat"]["id"])

    def get_message_id_from_message(self, message):
        """
        Messages containing multiple images have a `media_group_id` attribute.

        See also base class.
        """

        msg_id = message.get("media_group_id") or message.get("message_id")
        return str(msg_id)

    def get_message_text(self, message):
        """See base class."""

        # Telegram parses the text of a message before sending it.
        # The result can be found in the message's "entities".
        return {
            "text": message.get("text") or message.get("caption") or "",
            "entities": message.get("entities", []),
        }

    def process_text(self, live_post, message_text):
        """
        Use the message entities to convert links.

        A raw link isn't converted by Telegram and a link with a description
        is sent as a `text_link` entity.

        See also base class.
        """

        text = message_text["text"]
        len_text = len(text)
        entities = message_text["entities"]

        # Process the entities in reversed order to be able to edit the text in place.
        for entity in reversed(entities):
            url = ""
            start = entity["offset"]
            end = start + entity["length"]

            if entity["type"] == "url":
                url = description = text[start:end]

                if is_embed(url):
                    # Check if this can match an embed block, if so no conversion happens.
                    # It matches an embed block if it has a line in the text for itself.
                    if end == len_text or text[end] == "\n":
                        if start == 0 or text[start - 1] == "\n":
                            # This is an embed block, skip to the next entity
                            continue

            if entity["type"] == "text_link":
                url = entity["url"]
                description = text[start:end]

            if url:
                url = format_url(url)
                link = f'<a href="{url}">{description}</a>'
                text = text[:start] + link + text[end:]

        return super().process_text(live_post=live_post, message_text=text)

    def get_file_path(self, file_id):
        """
        Retrieves the `file_path` of a Telegram file.

        The `file_path` is necessary to have more infos about the image and to download it.

        Args:
            file_id (str): Id of the file to download.

        Returns:
            str: The `file_path` property of the file as sent by Telegram.
        """

        response = requests.get(
            get_base_telegram_url() + "getFile", params={"file_id": file_id}
        )
        return response.json()["result"]["file_path"]

    def get_message_files(self, message):
        """See base class."""

        if "photo" in message:
            # Choose original photo which is the last of the list
            photo = message["photo"][-1]
            photo["file_path"] = self.get_file_path(file_id=photo["file_id"])
            return [photo]
        return []

    def get_image_title(self, image):
        """See base class."""

        return image["file_path"].split("/")[-1].split(".")[0]

    def get_image_name(self, image):
        """See base class."""

        return image["file_path"].split("/")[-1]

    def get_image_mimetype(self, image):
        """See base class."""

        mimetype = image["file_path"].split("/")[-1].split(".")[-1]
        return "jpeg" if mimetype == "jpg" else mimetype

    def get_image_content(self, image):
        """See base class."""

        file_path = image["file_path"]
        response = requests.get(
            f"https://api.telegram.org/file/bot{get_telegram_bot_token()}/{file_path}"
        )
        return ContentFile(response.content)

    def get_image_dimensions(self, image):
        """See base class."""

        return image["width"], image["height"]

    def get_message_id_from_edited_message(self, message):
        """See base class."""

        return self.get_message_id_from_message(message=message)

    def get_message_text_from_edited_message(self, message):
        """See base class."""

        return self.get_message_text(message=message)

    def get_message_files_from_edited_message(self, message):
        """See base class."""

        return self.get_message_files(message=message)



from functools import lru_cache

from django.conf import settings


@lru_cache(maxsize=1)
def get_redis_url():
    return getattr(settings, "WAGTAIL_LIVE_REDIS_URL", "redis://127.0.0.1:6379/1")


@lru_cache(maxsize=1)
def get_live_server_host():
    return getattr(settings, "WAGTAIL_LIVE_SERVER_HOST", "localhost")


@lru_cache(maxsize=1)
def get_live_server_port():
    return getattr(settings, "WAGTAIL_LIVE_SERVER_PORT", 8765)



import asyncio

import websockets

from ..redis import RedisBus, make_channel_group_name
from ..utils import get_live_server_host, get_live_server_port, get_redis_url


# Define the broadcast method to be used by the bus.
async def broadcast(message, recipients):
    await asyncio.wait([ws.send(message) for ws in recipients])


class WebsocketsPublisherApp:
    bus = RedisBus(url=get_redis_url(), broadcast=broadcast)

    async def __call__(self):
        """Called once per session."""

        host = get_live_server_host()
        port = get_live_server_port()
        async with websockets.serve(self.handler, host, port):
            await self.bus.run()  # Run the bus forever

    async def handler(self, websocket, path):
        """
        Called once per new connection.

        Adds/removes the websocket connection to/from the channel group
        corresponding to the channel id found in the request's path.
        """

        channel_id = path.split("/")[-2]
        channel_name = make_channel_group_name(channel_id)

        await self.bus.subscribe(channel_name, websocket)

        try:
            await websocket.wait_closed()

        finally:
            await self.bus.unsubscribe(channel_name, websocket)


app = WebsocketsPublisherApp()



from functools import lru_cache

from wagtail_live.utils import get_setting_or_raise


@lru_cache(maxsize=1)
def get_telegram_bot_token():
    """
    Retrieves user's telegram bot token.

    Returns:
        str: telegram bot token.

    Raises:
        ImproperlyConfigured: if the telegram bot token isn't specified in settings.
    """

    return get_setting_or_raise("TELEGRAM_BOT_TOKEN", "Telegram bot token")


@lru_cache(maxsize=1)
def get_telegram_webhook_url():
    """
    Retrieves the webhook url which Telegram sends new updates to.

    Returns:
        str: a URL.

    Raises:
        ImproperlyConfigured: if the telegram webhook url isn't specified in settings.
    """

    base_webhook_url = get_setting_or_raise(
        "TELEGRAM_WEBHOOK_URL", "Telegram webhook URL"
    )

    if base_webhook_url.endswith("/"):
        base_webhook_url = base_webhook_url[:-1]

    return (
        f"{base_webhook_url}/wagtail_live/telegram/events/{get_telegram_bot_token()}/"
    )


@lru_cache(maxsize=1)
def get_base_telegram_url():
    """Returns the base URL to use when calling Telegram's API."""

    return f"https://api.telegram.org/bot{get_telegram_bot_token()}/"


def format_url(url):
    """Fixes telegram url format"""

    # Prevent e.g. www.example.com from being interpreted as relative url
    if not url.startswith("http") and not url.startswith("//"):
        url = f"//{url}"

    return url



""" Webapp URLs """

from django.urls import include, path

from . import views

api_patterns = [
    path("channels/", views.create_channel_view),
    path("channels/<slug:channel_name>/", views.delete_channel_view),
    path("messages/", views.create_message_view),
    path("messages/<int:pk>/", views.message_detail_view),
    path("images/<int:pk>/", views.delete_image_view),
]

urlpatterns = [
    path("api/", include(api_patterns)),
    path("channels/", views.channels_list_view, name="channels"),
    path(
        "channels/<slug:channel_name>/",
        views.channel_detail_view,
        name="channel_detail",
    ),
]



from asgiref.sync import async_to_sync
from channels.layers import get_channel_layer

from wagtail_live.publishers.websocket import BaseWebsocketPublisher


class DjangoChannelsPublisher(BaseWebsocketPublisher):
    """Django channels publisher."""

    def publish(self, channel_id, renders, removals):
        """Sends updates to the room group corresponding to channel_id."""

        channel_layer = get_channel_layer()
        group_name = f"liveblog_{channel_id}"
        message = {
            "type": "update",
            "renders": renders,
            "removals": removals,
        }

        async_to_sync(channel_layer.group_send)(group_name, message)



from .publisher import PieSocketPublisher
from .utils import get_piesocket_api_key, get_piesocket_endpoint, get_piesocket_secret

__all__ = [
    "PieSocketPublisher",
    "get_piesocket_api_key",
    "get_piesocket_endpoint",
    "get_piesocket_secret",
]



class BaseWebsocketPublisher:
    """Base class for publishers using the websocket technique."""

    def __call__(self, sender, channel_id, renders, removals, **kwargs):
        """
        Listens to the `live_page_update` signal.

        Args:
            sender (LivePageMixin):
                Sender of the signal.
            channel_id (str):
                ID of the channel corresponding to the updated page.
            renders (dict):
                Dict containing the new posts and the edited posts of the updated page.
            removals (list):
                List containing the id of the deleted posts for the updated page.
        """

        renders = {
            post.id: {
                "show": post.value["show"],
                "content": post.render(context={"block_id": post.id}),
            }
            for post in renders
        }
        return self.publish(channel_id=channel_id, renders=renders, removals=removals)

    def publish(self, channel_id, renders, removals):
        """
        Sends a new update:

        - to the `channel_id` "group" for django channels,
        - to an event bus for separate websocket servers,
        - to the service itself for hosted websocket services.

        Args:
            channel_id (str):
                ID of the channel corresponding to the updated page.
            renders (dict):
                Dict containing the new posts and the edited posts of the updated page.
            removals (list):
                List containing the id of the deleted posts for the updated page.
        """

        raise NotImplementedError



from functools import lru_cache

from wagtail_live.utils import get_setting_or_raise


@lru_cache(maxsize=1)
def get_piesocket_api_key():
    """
    Retrieves user's Piesocket API key.

    Returns:
        (str) Piesocket API key.

    Raises:
        (ImproperlyConfigured) if the Piesocket API key isn't specified in settings.
    """

    return get_setting_or_raise(
        setting="PIESOCKET_API_KEY", setting_str="PieSocket API Key"
    )


@lru_cache(maxsize=1)
def get_piesocket_secret():
    """
    Retrieves user's piesocket secret.

    Returns:
        (str) Piesocket secret.

    Raises:
        (ImproperlyConfigured) if the piesocket secret isn't specified in settings.
    """

    return get_setting_or_raise(
        setting="PIESOCKET_SECRET", setting_str="PieSocket secret"
    )


@lru_cache(maxsize=1)
def get_piesocket_endpoint():
    """
    Retrieves user's piesocket endpoint.

    Returns:
        (str) Piesocket endpoint.

    Raises:
        (ImproperlyConfigured) if the piesocket endpoint isn't specified in settings.
    """

    return get_setting_or_raise(
        setting="PIESOCKET_ENDPOINT", setting_str="PieSocket endpoint"
    )



from setuptools import find_packages, setup

from src.wagtail_live import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

install_requires = [
    "wagtail>=2.12",
]

docs_requires = [
    "mkdocs>=1.2,<1.3",
    "mkdocs-material>=7.2,<7.3",
    "mkdocstrings>=0.15.2,<0.16",
]

test_requires = [
    "black==21.5b2",
    "isort==5.8.0",
    "flake8==3.9.2",
    "pytest>=6.2,<6.3",
    "pytest-cov>=2.12,<3",
    "pytest-django>=4.3.0,<5",
    "pytest-factoryboy>=2.1.0,<3",
    "pytest-mock>=3.6.1,<3.7.0",
    "pytest-asyncio>=0.15.1,<0.16",
    "channels>=3.0.0,<4.0.0",
    "aioredis>=2.0.0,<3",
    "starlette>=0.16.0,<0.17",
    "websockets>=9.0,<10",
    "mock>=4.0.3,<5.0.0",
    "wagtail-factories>=2.0.1,<3",
]

build_requires = [
    "twine",
    "check-wheel-contents",
]

setup(
    name="wagtail-live",
    version=__version__,
    description="An app for high speed content publishing from a messaging app to a Wagtail site.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Tidiane Dia and Wagtail Core Team",
    author_email="hello@wagtail.io",
    url="https://github.com/wagtail/wagtail-live/",
    project_urls={
        "Documentation": "https://wagtail.github.io/wagtail-live/",
        "Source": "https://github.com/wagtail/wagtail-live/",
        "Issue tracker": "https://github.com/wagtail/wagtail-live/issues/",
    },
    install_requires=install_requires,
    tests_require=test_requires,
    extras_require={
        "test": test_requires,
        "docs": docs_requires,
        "build": build_requires,
    },
    package_dir={"": "src"},
    packages=find_packages("src"),
    include_package_data=True,
    zip_safe=False,
    license="BSD-3-Clause",
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Framework :: Django",
        "Framework :: Wagtail",
        "Framework :: Wagtail :: 2",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
)



import json
import logging

import requests

from wagtail_live.publishers.websocket import BaseWebsocketPublisher

from .utils import get_piesocket_api_key, get_piesocket_secret

logger = logging.getLogger(__name__)

publish_url = "https://www.piesocket.com/api/publish"
headers = {"Content-Type": "application/json"}


class PieSocketPublisher(BaseWebsocketPublisher):
    """PieSocket publisher."""

    def publish(self, channel_id, renders, removals):
        """See base class."""

        payload = json.dumps(
            {
                "key": get_piesocket_api_key(),
                "secret": get_piesocket_secret(),
                "channelId": channel_id,
                "message": {"renders": renders, "removals": removals},
            }
        )

        response = requests.post(publish_url, headers=headers, data=payload)
        if not response.ok:
            logger.error("Failed publishing new update to PieSocket API.")



""" Block types and block constructors are defined in this module."""

from wagtail.admin.rich_text.converters.editor_html import EditorHTMLConverter
from wagtail.core import rich_text
from wagtail.core.blocks import (
    BooleanBlock,
    CharBlock,
    DateTimeBlock,
    RichTextBlock,
    StreamBlock,
    StructBlock,
    StructValue,
)
from wagtail.embeds.blocks import EmbedBlock, EmbedValue
from wagtail.images.blocks import ImageChooserBlock


class ContentBlock(StreamBlock):
    """A block that represents a live post content."""

    text = RichTextBlock(help_text="Text of the message")
    image = ImageChooserBlock(help_text="Image of the message")
    embed = EmbedBlock(help_text="URL of the embed message")


class LivePostBlock(StructBlock):
    """A generic block that maps to a message in a messaging app."""

    message_id = CharBlock(help_text="Message's ID")
    created = DateTimeBlock(
        required=False, help_text="Date and time of message creation"
    )
    modified = DateTimeBlock(
        required=False,
        help_text="Date and time of last update",
    )
    show = BooleanBlock(
        required=False,
        help_text="Indicates if this message is shown/hidden",
        default=True,
    )
    content = ContentBlock()

    class Meta:
        template = "wagtail_live/blocks/live_post.html"


def construct_text_block(text):
    """
    Helper function to construct a text block for a LivePostBlock content.

    Args:
        text (str): Text to add

    Returns:
        RichText: a TextBlock filled with the given text.
    """

    # Make sure no malicious html is accepted
    # BeautifulSoup prefers markup that contains at least 1 tag,
    # if that's not the case we can accept the input as is.
    if "<" in text:
        features = rich_text.features.get_default_features()
        cleaned_text = EditorHTMLConverter(features=features).whitelister.clean(text)
    else:
        cleaned_text = text

    return RichTextBlock().to_python(cleaned_text)


def construct_image_block(image):
    """
    Helper function to construct an image block for a LivePostBlock content.

    Args:
        image (pk): Foreign key to the image to add

    Returns:
        ImageBlock: an ImageBlock filled with the given image.
    """

    return ImageChooserBlock().to_python(image.id)


def construct_embed_block(url):
    """
    Helper function to construct an embed block for a LivePostBlock content.

    Args:
        url (str): Url of the embed

    Returns:
        EmbedBlock: an EmbedBlock filled with the given url.
    """

    return EmbedBlock().to_python(url)


def construct_live_post_block(message_id, created):
    """
    Helper function to construct a LivePostBlock.

    Args:
        message_id (str):
            Id of the message to construct a live post for.
        created (DateTime):
            Date and time of message creation.

    Returns:
        LivePostBlock: a LivePostBlock
    """

    return LivePostBlock().to_python(
        {
            "message_id": message_id,
            "created": created,
        }
    )


def add_block_to_live_post(block_type, block, live_block):
    """
    Adds a new content block to a live post.

    Args:
        block_type (str):
            Type of the block to add
        block (Block):
            Block to add to the live post.
        live_block (LivePostBlock):
            Live post in which the new block will be added.
    """

    if isinstance(live_block, StructValue):
        live_block["content"].append((block_type, block))
    else:
        live_block.value["content"].append((block_type, block))


def clear_live_post_content(live_post):
    """
    Clears the content of a live post.

    Args:
        live_post (livePostBlock): Live post which content will be cleared.
    """

    if isinstance(live_post, StructValue):
        live_post["content"].clear()
    else:
        live_post.value["content"].clear()


def compare_live_posts_values(first_post_value, second_post_value):
    """
    Compares the values of two live posts.

    Args:
        first_post_value (StreamValue):   First live post.
        second_post_value (StreamValue):  Second live post.

    Returns:
        bool: True if the live posts are identic, False else.
    """

    if first_post_value["show"] != second_post_value["show"]:
        return False

    first_content = first_post_value["content"]
    second_content = second_post_value["content"]
    if len(first_content) != len(second_content):
        return False

    for i in range(0, len(first_content)):
        first_item = first_content[i]
        second_item = second_content[i]
        if (
            first_item.block_type != second_item.block_type
            or first_item.id != second_item.id
        ):
            return False

        first_value = first_item.value
        second_value = second_item.value
        if isinstance(first_value, rich_text.RichText):
            first_value = rich_text.get_text_for_indexing(first_value.source)
            second_value = rich_text.get_text_for_indexing(second_value.source)
        elif isinstance(first_value, EmbedValue):
            first_value = first_value.url
            second_value = second_value.url

        if first_value != second_value:
            return False

    return True



""" Webapp views """

from django.http import JsonResponse
from django.shortcuts import get_object_or_404
from django.utils.timezone import now
from django.views.generic import DetailView, ListView
from rest_framework import status
from rest_framework.renderers import TemplateHTMLRenderer
from rest_framework.response import Response
from rest_framework.views import APIView

from .models import Channel, Image, Message
from .receiver import MESSAGE_CREATED, MESSAGE_DELETED, MESSAGE_EDITED, WebAppReceiver
from .serializers import ChannelSerializer, MessageSerializer

LIVE_RECEIVER = WebAppReceiver()


def send_update(update_type, data):
    event = {"update_type": update_type}
    event.update(data)
    LIVE_RECEIVER.dispatch_event(event=event)


class ChannelListView(ListView):
    """List all channels"""

    model = Channel
    context_object_name = "channels"


channels_list_view = ChannelListView.as_view()


class ChannelDetailView(DetailView):
    """Channel details view"""

    model = Channel
    context_object_name = "channel"
    slug_field = "channel_name"
    slug_url_kwarg = "channel_name"


channel_detail_view = ChannelDetailView.as_view()


class CreateChannelView(APIView):
    def post(self, request):
        """API endpoint: create a new channel"""

        serializer = ChannelSerializer(data=request.data)
        if serializer.is_valid():
            serializer.save()
            return Response(serializer.data, status=status.HTTP_201_CREATED)
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)


create_channel_view = CreateChannelView.as_view()


class DeleteChannelView(APIView):
    slug_field = "channel_name"
    slug_url_kwarg = "channel_name"

    def delete(self, request, channel_name):
        """API endpoint: delete a channel by its name"""

        channel = get_object_or_404(Channel, channel_name=channel_name)
        channel.delete()
        return Response(status=status.HTTP_204_NO_CONTENT)


delete_channel_view = DeleteChannelView.as_view()


class CreateMessageView(APIView):
    renderer_classes = [TemplateHTMLRenderer]
    template_name = "webapp/message.html"

    def post(self, request):
        """API endpoint: create a new message"""
        serializer = MessageSerializer(data=request.data)
        if serializer.is_valid():
            serializer.save()
            data = serializer.data
            send_update(update_type=MESSAGE_CREATED, data=data)
            return Response({"message": data}, status=status.HTTP_201_CREATED)
        return JsonResponse(serializer.errors, status=status.HTTP_400_BAD_REQUEST)


create_message_view = CreateMessageView.as_view()


class MessageDetailView(APIView):
    """
    Retrieve, update or delete a Message instance.
    """

    renderer_classes = [TemplateHTMLRenderer]
    template_name = "webapp/message.html"

    def put(self, request, pk):
        """API endpoint: update a message by its ID"""

        message = get_object_or_404(Message, pk=pk)
        serializer = MessageSerializer(message, data=request.data)
        if serializer.is_valid():
            serializer.validated_data["modified"] = now()
            serializer.save()
            data = serializer.data
            send_update(update_type=MESSAGE_EDITED, data=data)
            return Response({"message": data})
        return JsonResponse(data=serializer.errors, status=status.HTTP_400_BAD_REQUEST)

    def delete(self, request, pk):
        """API endpoint: delete a message by its ID"""

        message = get_object_or_404(Message, pk=pk)
        update = {
            "channel": message.channel.channel_name,
            "id": pk,
        }
        message.delete()
        send_update(update_type=MESSAGE_DELETED, data=update)
        return JsonResponse(data={}, status=status.HTTP_204_NO_CONTENT)


message_detail_view = MessageDetailView.as_view()


class DeleteImageView(APIView):
    def delete(self, request, pk):
        """API endpoint: delete an image by its ID"""

        image = get_object_or_404(Image, pk=pk)
        image.delete()
        return Response(status=status.HTTP_204_NO_CONTENT)


delete_image_view = DeleteImageView.as_view()



from channels.generic.websocket import AsyncJsonWebsocketConsumer
from django.urls import re_path


class DjangoChannelsApp(AsyncJsonWebsocketConsumer):
    """App/Consumer which handles websocket connections."""

    async def connect(self):
        """Adds websocket channel to room group."""

        self.channel_id = self.scope["url_route"]["kwargs"]["channel_id"]
        self.group_name = f"liveblog_{self.channel_id}"

        await self.channel_layer.group_add(self.group_name, self.channel_name)
        await self.accept()

    async def disconnect(self, close_code):
        """Discards websocket channel from room group."""

        await self.channel_layer.group_discard(self.group_name, self.channel_name)

    async def update(self, event):
        """Receives messages from room group and sends them to websocket client."""

        await self.send_json(
            {"renders": event["renders"], "removals": event["removals"]}
        )


live_websocket_route = [
    re_path(r"ws/channel/(?P<channel_id>\w+)/$", DjangoChannelsApp.as_asgi()),
]



# Generated by Django 3.2 on 2021-07-07 17:39

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Channel",
            fields=[
                (
                    "channel_name",
                    models.CharField(
                        help_text="Channel name",
                        max_length=124,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        auto_now_add=True, help_text="Date and time of channel creation"
                    ),
                ),
            ],
            options={
                "ordering": ["-created"],
            },
        ),
        migrations.CreateModel(
            name="Message",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        auto_now_add=True, help_text="Date and time of message creation"
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True, help_text="Date and time of last update", null=True
                    ),
                ),
                (
                    "show",
                    models.BooleanField(
                        blank=True,
                        default=True,
                        help_text="Indicates if this message is shown/hidden",
                    ),
                ),
                ("content", models.TextField(help_text="Content of the message")),
                (
                    "channel",
                    models.ForeignKey(
                        help_text="Channel this message was posted to.",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="messages",
                        to="webapp.channel",
                    ),
                ),
            ],
            options={
                "ordering": ["-created"],
            },
        ),
        migrations.CreateModel(
            name="Image",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "image",
                    models.ImageField(help_text="Image of a message", upload_to=""),
                ),
                (
                    "message",
                    models.ForeignKey(
                        help_text="Message this image was posted to.",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="files",
                        to="webapp.message",
                    ),
                ),
            ],
        ),
    ]
